@@ -16,8 +16,14 @@
  *
  *	Added IPv6 support.
  *
+ * Updated: Joshua Brindle <jbrindle@tresys.com>
+ *	    Karl MacMillan <kmacmillan@tresys.com>
+ *          Jason Tang     <jtang@tresys.com>
+ *
+ *	Policy Module support.
+ *
  * Copyright (C) 2004-2005 Trusted Computer Solutions, Inc.
- * Copyright (C) 2003 - 2004 Tresys Technology, LLC
+ * Copyright (C) 2003 - 2005 Tresys Technology, LLC
  * Copyright (C) 2003 Red Hat, Inc., James Morris <jmorris@redhat.com>
  *	This program is free software; you can redistribute it and/or modify
  *  	it under the terms of the GNU General Public License as published by
@@ -64,11 +70,12 @@
 #include <errno.h>
 #include <sys/mman.h>
 
-#include <sepol/policydb.h>
-#include <sepol/services.h>
-#include <sepol/conditional.h>
-#include <sepol/hierarchy.h>
-#include <sepol/flask.h>
+#include <sepol/policydb/policydb.h>
+#include <sepol/policydb/services.h>
+#include <sepol/policydb/conditional.h>
+#include <sepol/policydb/hierarchy.h>
+#include <sepol/policydb/flask.h>
+#include <sepol/policydb/expand.h>
 
 #include "queue.h"
 #include "checkpolicy.h"
@@ -83,19 +90,18 @@ extern policydb_t *policydbp;
 extern queue_t id_queue;
 extern unsigned int policydb_errors;
 extern unsigned long policydb_lineno;
-extern unsigned long source_lineno;
 extern char source_file[];
-extern unsigned int pass;
+extern int mlspol;
 
 extern FILE *yyin;
+extern void init_parser(int);
 extern int yyparse(void);
 extern void yyrestart(FILE *);
 
-char *txtfile = "policy.conf";
-char *binfile = "policy";
+static char *txtfile = "policy.conf";
+static char *binfile = "policy";
 
 unsigned int policyvers = POLICYDB_VERSION_MAX;
-unsigned int mlspol = 0;
 
 void usage(char *progname)
 {
@@ -147,16 +153,6 @@ static int find_perm(hashtab_key_t key, hashtab_datum_t datum, void *p)
 	return 0;
 }
 
-static int type_attr_remove(hashtab_key_t key __attribute__ ((unused)), hashtab_datum_t datum, void *p __attribute__ ((unused)))
-{
-	type_datum_t *typdatum;
-
-	typdatum = (type_datum_t *) datum;
-	if (typdatum->isattr)
-		return 1;
-	return 0;
-}
-
 #ifdef EQUIVTYPES
 static int insert_type_rule(avtab_key_t *k, avtab_datum_t *d, 
 			    struct avtab_node *type_rules)
@@ -284,99 +280,8 @@ static int identify_equiv_types(void)
 }
 #endif 
 
-static void cond_check_type_rules_list(cond_av_list_t *list)
-{
-	cond_av_list_t *cur = list;
-
-	while (cur) {
-		if (cur->node->datum.specified & AVTAB_TYPE) {
-			if (avtab_search(&policydbp->te_avtab, &cur->node->key, AVTAB_TYPE)) {
-				fprintf(stderr, "conditional type rule for (%s, %s : %s) conflicts with entry in base policy; "
-					"conditional rule discarded.\n", policydbp->p_type_val_to_name[cur->node->key.source_type-1],
-					policydbp->p_type_val_to_name[cur->node->key.target_type-1],
-					policydbp->p_class_val_to_name[cur->node->key.target_class-1]);
-				cur->node->parse_context = (void*)0;
-			} else {
-				cur->node->parse_context = (void*)1;
-			}
-		} else {
-			cur->node->parse_context = (void*)1;
-		}
-		cur = cur->next;
-	}
-}
-
-/* check for duplicate type rules - this has to be done after all of
- * the parsing is finished because the conditional and base type rules
- * are collected in the same pass */
-static void cond_check_type_rules(void)
-{
-	cond_list_t *node;
-
-	for (node = policydbp->cond_list; node != NULL; node = node->next) {
-		cond_check_type_rules_list(node->true_list);
-		cond_check_type_rules_list(node->false_list);
-	}
-}
-
 extern char *av_to_string(uint32_t tclass, sepol_access_vector_t av);
 
-void check_assertion_helper(unsigned int stype, unsigned int ttype, ebitmap_t *tclasses, 
-			    sepol_access_vector_t *avp, unsigned long line) 
-{
-	avtab_key_t avkey;
-	avtab_datum_t *avdatump;
-	unsigned int k;
-	
-
-	for (k = ebitmap_startbit(tclasses); k < ebitmap_length(tclasses); k++) {
-		if (!ebitmap_get_bit(tclasses, k)) 
-			continue;
-		avkey.source_type = stype + 1;
-		avkey.target_type = ttype + 1;
-		avkey.target_class = k + 1;
-		avdatump = avtab_search(&policydb.te_avtab, &avkey, AVTAB_AV);
-		if (!avdatump) 
-			continue;
-
-		if ((avdatump->specified & AVTAB_ALLOWED) &&
-		    (avtab_allowed(avdatump) & avp[k])) {
-			fprintf(stderr, "assertion on line %ld violated by allow %s %s:%s {%s };\n", line, policydb.p_type_val_to_name[stype], policydb.p_type_val_to_name[ttype], policydb.p_class_val_to_name[k], 
-				av_to_string(k+1,
-					     avtab_allowed(avdatump) & avp[k]));
-			policydb_errors++;
-		}
-	}
-}
-
-void check_assertions(void) 
-{
-	te_assert_t *a, *tmp;
-	unsigned int i, j;
-
-	a = te_assertions;
-	while (a) {
-		for (i = ebitmap_startbit(&a->stypes); i < ebitmap_length(&a->stypes); i++) {
-			if (!ebitmap_get_bit(&a->stypes, i))
-				continue;
-			if (a->self) {
-				check_assertion_helper(i, i, &a->tclasses, a->avp, a->line);
-			}
-			for (j = ebitmap_startbit(&a->ttypes); j < ebitmap_length(&a->ttypes); j++) {
-				if (!ebitmap_get_bit(&a->ttypes, j)) 
-					continue;
-				check_assertion_helper(i, j, &a->tclasses, a->avp, a->line);
-			}
-		}
-		tmp = a;
-		a = a->next;
-		ebitmap_destroy(&tmp->stypes);
-		ebitmap_destroy(&tmp->ttypes);
-		ebitmap_destroy(&tmp->tclasses);
-		free(tmp->avp);
-		free(tmp);
-	}
-}
 
 int display_bools()
 {
@@ -456,7 +361,7 @@ int main(int argc, char **argv)
 	sepol_security_context_t scontext;
 	struct sepol_av_decision avd;
 	class_datum_t *cladatum;
-	char ans[80 + 1], *file = txtfile, *outfile = NULL, *path, *fstype, error_msg[ERRMSG_LEN];
+	char ans[80 + 1], *file = txtfile, *outfile = NULL, *path, *fstype;
 	size_t scontext_len, pathlen;
 	unsigned int i;
 	unsigned int protocol, port;
@@ -553,6 +458,10 @@ int main(int argc, char **argv)
 		pf.type = PF_USE_MEMORY;
 		pf.data = map;
 		pf.len = sb.st_size;
+		if (policydb_init(&policydb)) {
+			fprintf(stderr, "%s:  policydb_init:  Out of memory!\n", argv[0]);
+			exit(1);
+		}
 		ret = policydb_read(&policydb, &pf, 1);
 		if (ret) {
 			fprintf(stderr, "%s:  error(s) encountered while parsing configuration\n", argv[0]);
@@ -561,7 +470,7 @@ int main(int argc, char **argv)
 		policydbp = &policydb;
 
 		/* Check Policy Consistency */
-		if (sepol_mls_enabled()) {
+		if (policydbp->mls) {
 			if (!mlspol) {
 				fprintf(stderr,"%s:  MLS policy, but non-MLS"
 				        " is specified\n", argv[0]);
@@ -575,65 +484,58 @@ int main(int argc, char **argv)
 			}
 		}
 	} else {
-		yyin = fopen(file, "r");
+                policydb_t parse_policy;
+
+                yyin = fopen(file, "r");
 		if (!yyin) {
 			fprintf(stderr, "%s:  unable to open %s\n", argv[0], 
 				file);
 			exit(1);
 		}
 
-		if (policydb_init(&policydb))
+		if (policydb_init(&parse_policy))
 			exit(1);
+		/* We build this as a base policy first since that is all the parser understands */
+		parse_policy.policy_type = POLICY_BASE;
 
 		/* Let sepol know if we are dealing with MLS support */
-		sepol_set_mls(mlspol);
+		parse_policy.mls = mlspol;
 
 		id_queue = queue_create();
 		if (!id_queue) {
 			fprintf(stderr, "%s:  out of memory\n", argv[0]);
 			exit(1);
 		}
-		policydbp = &policydb;
-		policydb_errors = 0;
-		pass = 1;
+		policydbp = &parse_policy;
+                init_parser(1);
 		if (yyparse() || policydb_errors) {
 			fprintf(stderr, "%s:  error(s) encountered while parsing configuration\n", argv[0]);
 			exit(1);
 		}
 		rewind(yyin);
-		policydb_lineno = 1;
+                init_parser(2);
 		source_file[0] = '\0';
-		source_lineno = 1;
 		yyrestart(yyin);
-		pass = 2;
 		if (yyparse() || policydb_errors) {
 			fprintf(stderr, "%s:  error(s) encountered while parsing configuration\n", argv[0]);
 			exit(1);
 		}
 		queue_destroy(id_queue);
 
-		cond_check_type_rules();
-		cond_optimize_lists(policydb.cond_list);
-
-		check_assertions();
 		if (policydb_errors) 
 			exit(1);
 
-		if (policyvers >= POLICYDB_VERSION_NLCLASS &&
-		    policydb.p_classes.nprim < SECCLASS_NETLINK_DNRT_SOCKET) {
-			fprintf(stderr, "%s:  policy lacks new netlink classes, unable to generate policy version %d\n", argv[0], policyvers);
+		if (policydb_init(&policydb)) {
+			fprintf(stderr, "%s:  policydb_init failed\n", argv[0]);
 			exit(1);
 		}
 
-		if (hierarchy_check_constraints(&policydb, error_msg, sizeof(error_msg))) {
-			fprintf(stderr, "%s\n", error_msg);
+		if (expand_module(NULL, &parse_policy, &policydb, 0, 1)) {
+                        fprintf(stderr, "Error while expanding policy\n");
 			exit(1);
-		}
-
-		/* remove type attributes */
-		hashtab_map_remove_on_error(policydb.p_types.table, 
-					    type_attr_remove, 0, 0);
-		
+                }
+		policydb_destroy(&parse_policy);
+		policydbp = &policydb;
 		fclose(yyin);
 	}
 
@@ -651,12 +553,8 @@ int main(int argc, char **argv)
 			exit(1);
 		}
 
-		if (sepol_set_policyvers(policyvers)) {
-			fprintf(stderr,"%s:  incompatible policy (version %d) "
-				"for writing to %s\n", argv[0],
-				policyvers, outfile);
-			exit(1);
-		}
+		policydb.policy_type = POLICY_KERN;
+		policydb.policyvers = policyvers;
 
 		pf.type = PF_USE_STDIO;
 		pf.fp = outfp;
@@ -668,8 +566,10 @@ int main(int argc, char **argv)
 		}
 		fclose(outfp);
 	}
-	if (!debug)
+	if (!debug) {
+		policydb_destroy(&policydb);
 		exit(0);
+	}
 
       menu:
 	printf("\nSelect an option:\n");
