@@ -1,7 +1,7 @@
 #ifndef _CHECKPOLICY_H_
 #define _CHECKPOLICY_H_
 
-#include <sepol/ebitmap.h>
+#include <sepol/policydb/ebitmap.h>
 
 typedef struct te_assert { 
 	ebitmap_t stypes;
@@ -16,6 +16,5 @@ typedef struct te_assert {
 te_assert_t *te_assertions;
 
 extern unsigned int policyvers;
-extern unsigned int mlspol;
 
 #endif
