@@ -17,10 +17,11 @@
  * 	displaypol binary_pol_file
  */
 
-#include <sepol/policydb.h>
-#include <sepol/avtab.h>
-#include <sepol/services.h>
-#include <sepol/conditional.h>
+#include <sepol/policydb/policydb.h>
+#include <sepol/policydb/avtab.h>
+#include <sepol/policydb/services.h>
+#include <sepol/policydb/conditional.h>
+#include <sepol/policydb/expand.h>
 #include <getopt.h>
 #include <assert.h>
 #include <unistd.h>
@@ -40,11 +41,6 @@ void usage(char *progname)
 	exit(1);
 }
 
-/* borrowed from avtab.c */
-#define HASH_BITS 15
-#define HASH_BUCKETS (1 << HASH_BITS)
-#define HASH_MASK (HASH_BUCKETS-1)
-
 /* borrowed from checkpolicy.c */
 static int find_perm(hashtab_key_t key, hashtab_datum_t datum, void *p)
 {
@@ -92,9 +88,18 @@ int render_type(uint32_t type, policydb_t *p, FILE *fp)
 
 int render_key(avtab_key_t *key, policydb_t *p, FILE *fp)
 {
-	fprintf(fp, "%s %s : %s ", p->p_type_val_to_name[key->source_type - 1], 
-			p->p_type_val_to_name[key->target_type - 1],
-			p->p_class_val_to_name[key->target_class - 1]);
+	char *stype, *ttype, *tclass;
+	stype = p->p_type_val_to_name[key->source_type - 1];
+	ttype = p->p_type_val_to_name[key->target_type - 1];
+	tclass = p->p_class_val_to_name[key->target_class - 1];
+	if (stype && ttype) 
+		fprintf(fp, "%s %s : %s ", stype, ttype, tclass);
+	else if (stype)
+		fprintf(fp, "%s %u : %s ", stype, key->target_type, tclass);
+	else if (ttype)
+		fprintf(fp, "%u %s : %s ", key->source_type, ttype, tclass);
+	else
+		fprintf(fp, "%u %u : %s ", key->source_type, key->target_type, tclass);
 	return 0;
 }
 
@@ -108,58 +113,58 @@ int render_av_rule(avtab_key_t *key, avtab_datum_t *datum, uint32_t what, policy
 {
 	if(!(what & RENDER_UNCONDITIONAL)) {
 		if(what != RENDER_CONDITIONAL &&
-		   (((what & RENDER_ENABLED) && !(datum->specified & AVTAB_ENABLED)) ||
-		   ((what & RENDER_DISABLED) && (datum->specified & AVTAB_ENABLED)))) {
+		   (((what & RENDER_ENABLED) && !(key->specified & AVTAB_ENABLED)) ||
+		   ((what & RENDER_DISABLED) && (key->specified & AVTAB_ENABLED)))) {
 		   	return 0; /* doesn't match selection criteria */
 		} 
 	}
 
 	if(!(what & RENDER_UNCONDITIONAL)) {
-		if(datum->specified & AVTAB_ENABLED) 
+		if(key->specified & AVTAB_ENABLED) 
 			fprintf(fp, "[enabled] ");
-		else if(!(datum->specified & AVTAB_ENABLED))
+		else if(!(key->specified & AVTAB_ENABLED))
 			fprintf(fp, "[disabled] ");
 	} 
 
-	if( datum->specified & AVTAB_AV) {
-		if(datum->specified & AVTAB_ALLOWED) {
+	if( key->specified & AVTAB_AV) {
+		if(key->specified & AVTAB_ALLOWED) {
 			fprintf(fp, "allow ");			
 			render_key(key, p, fp);
-			render_access_mask(avtab_allowed(datum),key, p, fp);
+			render_access_mask(datum->data,key, p, fp);
 			fprintf(fp, ";\n");
 		}
-		if(datum->specified & AVTAB_AUDITALLOW) {
+		if(key->specified & AVTAB_AUDITALLOW) {
 			fprintf(fp, "auditallow ");
 			render_key(key, p, fp);
-			render_access_mask(avtab_auditallow(datum),key, p, fp);
+			render_access_mask(datum->data,key, p, fp);
 			fprintf(fp, ";\n");
 		}
-		if(datum->specified & AVTAB_AUDITDENY) {
+		if(key->specified & AVTAB_AUDITDENY) {
 			fprintf(fp, "dontaudit ");
 			render_key(key, p, fp);
 			/* We inverse the mask for dontaudit since the mask is internally stored
 			 * as a auditdeny mask */
-			render_access_mask(~avtab_auditdeny(datum),key, p, fp);
+			render_access_mask(~datum->data,key, p, fp);
 			fprintf(fp, ";\n");
 		}
 	}
-	else if( datum->specified & AVTAB_TYPE){
-		if(datum->specified & AVTAB_TRANSITION) {
+	else if( key->specified & AVTAB_TYPE){
+		if(key->specified & AVTAB_TRANSITION) {
 			fprintf(fp, "type_transition ");
 			render_key(key, p, fp);
-			render_type(avtab_transition(datum), p, fp);
+			render_type(datum->data, p, fp);
 			fprintf(fp, ";\n");
 		}
-		if(datum->specified & AVTAB_MEMBER) {
+		if(key->specified & AVTAB_MEMBER) {
 			fprintf(fp, "type_member ");
 			render_key(key, p, fp);
-			render_type(avtab_member(datum), p, fp);
+			render_type(datum->data, p, fp);
 			fprintf(fp, ";\n");
 		}
-		if(datum->specified & AVTAB_CHANGE) {
+		if(key->specified & AVTAB_CHANGE) {
 			fprintf(fp, "type_change ");
 			render_key(key, p, fp);
-			render_type(avtab_change(datum), p, fp);
+			render_type(datum->data, p, fp);
 			fprintf(fp, ";\n");
 		}
 	}
@@ -174,14 +179,27 @@ int display_avtab(avtab_t *a, uint32_t what, policydb_t *p, FILE *fp)
 {
 	int i;
 	avtab_ptr_t cur;
+	avtab_t expa;
 
+	if (avtab_init(&expa)) 
+		goto oom;
+	if (expand_avtab(p, a, &expa)) {
+		avtab_destroy(&expa);
+		goto oom;
+	}
+
+	/* hmm...should have used avtab_map. */
 	for (i = 0; i < AVTAB_SIZE; i++) {
-		for (cur = a->htable[i]; cur; cur = cur->next) {
+		for (cur = expa.htable[i]; cur; cur = cur->next) {
 			render_av_rule(&cur->key, &cur->datum, what, p, fp);
 		}
 	}
+	avtab_destroy(&expa);
 	fprintf(fp, "\n");
 	return 0;
+oom:
+	fprintf(stderr, "out of memory\n");
+	return 1;
 }
 
 int display_bools(policydb_t *p, FILE *fp)
@@ -232,25 +250,48 @@ void display_expr(policydb_t *p, cond_expr_t *exp, FILE *fp)
 int display_cond_expressions(policydb_t *p, FILE *fp)
 {
 	cond_node_t *cur;
-	cond_av_list_t *av_cur;
+	cond_av_list_t *av_cur, *expl = NULL;
+	avtab_t expa;
+
+
 	for (cur = p->cond_list; cur != NULL; cur = cur->next) {
 		fprintf(fp, "expression: ");
 		display_expr(p, cur->expr, fp);
 		fprintf(fp, "current state: %d\n", cur->cur_state);
 		fprintf(fp, "True list:\n");
-		for (av_cur = cur->true_list; av_cur != NULL; av_cur = av_cur->next) {
+		if (avtab_init(&expa))
+			goto oom;
+		if (expand_cond_av_list(p, cur->true_list, &expl, &expa)) {
+			avtab_destroy(&expa);
+			goto oom;
+		}
+		for (av_cur = expl; av_cur != NULL; av_cur = av_cur->next) {
 			fprintf(fp, "\t");
 			render_av_rule(&av_cur->node->key, &av_cur->node->datum,
 				       RENDER_CONDITIONAL, p, fp);
 		}
+		cond_av_list_destroy(expl);
+		avtab_destroy(&expa);
 		fprintf(fp, "False list:\n");
-		for (av_cur = cur->false_list; av_cur != NULL; av_cur = av_cur->next) {
+		if (avtab_init(&expa))
+			goto oom;
+		if (expand_cond_av_list(p, cur->false_list, &expl, &expa)) {
+			avtab_destroy(&expa);
+			goto oom;
+		}
+		for (av_cur = expl; av_cur != NULL; av_cur = av_cur->next) {
 			fprintf(fp, "\t");
 			render_av_rule(&av_cur->node->key, &av_cur->node->datum,
 				       RENDER_CONDITIONAL, p, fp);
 		}
+		cond_av_list_destroy(expl);
+		avtab_destroy(&expa);
 	}
 	return 0;
+
+oom:
+	fprintf(stderr, "out of memory\n");
+	return 1;
 }
 
 int change_bool(char *name, int state, policydb_t *p, FILE *fp)
@@ -322,6 +363,10 @@ int main(int argc, char **argv)
 	pf.type = PF_USE_MEMORY;
 	pf.data = map;
 	pf.len = sb.st_size;
+	if (policydb_init(&policydb)) {
+		fprintf(stderr, "%s:  Out of memory!\n", argv[0]);
+		exit(1);
+	}
 	ret = policydb_read(&policydb, &pf, 1);
 	if (ret) {
 		fprintf(stderr, "%s:  error(s) encountered while parsing configuration\n", argv[0]);
